////////////////////////////////////////////////////////////////////////////////
//  3  : "  "
//:  Ը 
//: http://www.DXGames.narod.ru
//e-mail: UrvanovF@hotmail.ru
////////////////////////////////////////////////////////////////////////////////
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <memory.h>

const int BLOCK_SIZE=8;
typedef  int BLOCK[BLOCK_SIZE]; 
enum CHANGESTRATEGY{csLRU,csRANDOM};
enum WRITESTRATEGY{wsTHROUGH,wsBACK};

const int OZUBlockCount=1024;
const int OZUSize=BLOCK_SIZE*OZUBlockCount;
const int CacheBlockCount=8;
const int CacheSize=CacheBlockCount*BLOCK_SIZE;

BLOCK * OZU=0,* Cache=0;
int CacheHandbook[CacheBlockCount];
int Oldness[CacheBlockCount];
CHANGESTRATEGY ChangeStrategy=csRANDOM;
WRITESTRATEGY WriteStrategy=wsTHROUGH;

////////////////////////////////////////////////////////////////////////////////
//  
int GetCount=0;
int MissCount=0;
int CallCount=0;
/////////////////////////////////////////////////
void Initialize();
void Deinitialize();
void Sort();
void WriteOZU(int Address,int Number);
int ReadOZU(int Address);
bool IsInCash(int addr, int * block);
void LoadBlock(int Address);
int Oldest();

int main(int argc, char *argv[])
{
	try{
		Initialize();
		Sort();
		printf("Get count: %i\n",GetCount);
		printf("Miss count: %i\n",MissCount);
		printf("Call count: %i\n",CallCount);
		printf("Percent missing: %3.2f\n",(float)MissCount/(float)CallCount*100.0);
		Deinitialize();
		return 0;
	}
	catch(char *s)
	{
		fprintf(stderr,"Error: %s",s);
		Deinitialize();
		return 1;
	}
}

void Initialize()
{
	int n,m;
	OZU=new BLOCK[OZUBlockCount];
	Cache= new BLOCK[CacheBlockCount];
	memset((void*)CacheHandbook,-1,CacheBlockCount);
	memset((void*)CacheHandbook,0,CacheBlockCount);

	srand((unsigned )time(0));
	for (n=0;n<OZUBlockCount;n++)
	{
		for (m=0;m<BLOCK_SIZE;m++)
		{
			OZU[n][m]=rand();
		}
	}
	if ((!OZU)||(!Cache))
	{
		throw "Not enough memory.";
	}
}

void Deinitialize()
{
	if (OZU)
	{
		delete []OZU;
		OZU=0;
	}
	if (Cache)
	{
		delete []Cache;
		Cache=0;
	}
}

void Sort()
{
	int n,m,v1,v2,buff;
	const int First=0, End=BLOCK_SIZE*OZUBlockCount,Size=End-First;
	printf("Calculating. Please wait...");
	for (n=First;n<End;n++)
	{
		for (m=n+1;m<End;m++)
		{
			v1=ReadOZU(n);
			v2=ReadOZU(m);
			if (v1>v2)
			{
				buff=v2;
				WriteOZU(m,v1);
				WriteOZU(n,buff);
			}
		}
		printf("Progress: %3.2f\n",((float)n)/((float)Size-1.0)*100.0);
	}
}

void WriteOZU(int Address,int Number)
{
	int block;
	if (WriteStrategy == wsTHROUGH)
	{
		OZU[Address/BLOCK_SIZE][Address%BLOCK_SIZE]=Number;
		if (IsInCash(Address,&block))
		{
			Cache[block][Address%BLOCK_SIZE] = Number;
			
		}
	} 
	else
	{
		if (IsInCash(Address,&block))
		{
			Cache[block][Address%BLOCK_SIZE] = Number;
		}
		else
		{
			OZU[Address/BLOCK_SIZE][Address%BLOCK_SIZE]=Number;
		}
	}
}

int ReadOZU(int Address)
{
	int block,res;
	CallCount++;
	if (IsInCash(Address,&block)) 
	{	
	   res=Cache[block][Address%BLOCK_SIZE];
	   GetCount++;
	   Oldness[block]=0;
	   
	}
	else 
	{
		res=OZU[Address/BLOCK_SIZE][Address%BLOCK_SIZE];
		MissCount++;
		LoadBlock(Address);
		
	}
	return res;
}

bool IsInCash(int addr, int * block)
{
	bool flag=false;
	for (int i=0;i< CacheBlockCount;i++)
		if (CacheHandbook[i] == addr/BLOCK_SIZE) {*block=i;flag=true; break;};
	 return flag;
}

////////////////////////////////////////////////////////////////////////////////
//   
////////////////////////////////////////////////////////////////////////////////
void LoadBlock(int Address)
{
	int Number=0;
	if (ChangeStrategy==csRANDOM)
	{
		Number = (rand()%CacheBlockCount);
	}
	else 
	{
		Number = Oldest();
		Oldness[Number]=0;
	}

	//    -   ( );
	if (WriteStrategy == wsBACK)
	{
		if (CacheHandbook[Number]!=-1) for (int n=0;n<BLOCK_SIZE;n++)
			OZU[CacheHandbook[Number]][n]=Cache[Number][n];
	};
	
	CacheHandbook[Number]=Address/BLOCK_SIZE;
	for (int n=0;n<BLOCK_SIZE;n++) 
	Cache[Number][n]=OZU[Address/BLOCK_SIZE][n];

	if (ChangeStrategy==csLRU)
		for (n=0;n<CacheBlockCount;n++) Oldness[n]++;
}

////////////////////////////////////////////////////////////////////////////////
//    ,      
////////////////////////////////////////////////////////////////////////////////
int Oldest()
{
	int res = 0;
	for (int i=1;i<CacheBlockCount;i++) 
		if ((Oldness[i]>Oldness[res])/*&&(Assotiation[i]!=0xffffffff)*/) res = i;
	return res;
};